'use strict';

var visit = require('unist-util-visit');
var is = require('unist-util-is');
var toString = require('nlcst-to-string');
var toWords = require('number-to-words').toWords;
var requiresA = require('./a');
var requiresAn = require('./an');

module.exports = indefiniteArticle;

var vowel = /[aeiou]/;
var digits = /^\d+/;
var join = /^(and|or|nor)$/i;
var punctuation = /^[“”‘’'"()[\]]$/;
var split = /['’ -]/;

requiresA = factory(requiresA);
requiresAn = factory(requiresAn);

function indefiniteArticle() {
  return transformer;

  function transformer(tree, file) {
    visit(tree, 'WordNode', visitor);

    function visitor(node, index, parent) {
      var value = toString(node);
      var normal = lower(value);
      var message;
      var following;
      var suggestion;
      var an;

      if (normal !== 'a' && normal !== 'an') {
        return;
      }

      an = value.length !== 1;
      following = after(parent, index);

      if (!following) {
        return;
      }

      following = toString(following);

      /* Exit if `A` and this isn’t sentence-start: `Station A equals` */
      if (normal !== value && !an && !firstWord(parent, index)) {
        return;
      }

      /* Exit if `a` is used as a letter: `a and b`. */
      if (normal === value && !an && join.test(following)) {
        return;
      }

      suggestion = classify(following);

      if ((suggestion === 'an' && !an) || (suggestion === 'a' && an)) {
        if (normal !== value) {
          suggestion = suggestion.charAt(0).toUpperCase() + suggestion.slice(1);
        }

        message = file.warn(
          'Use `' + suggestion + '` before `' + following + '`, ' +
          'not `' + value + '`',
          node
        );

        message.source = 'retext-indefinite-article';
        message.ruleId = 'retext-indefinite-article';
        message.actual = value;
        message.expected = [suggestion];
      }
    }
  }
}

/** Check if there’s no word before `index`. */
function firstWord(parent, index) {
  var siblings = parent.children;

  while (index--) {
    if (is('WordNode', siblings[index])) {
      return false;
    }
  }

  return true;
}

/** Get the next word. */
function after(parent, index) {
  var siblings = parent.children;
  var sibling = siblings[++index];
  var other;

  if (is('WhiteSpaceNode', sibling)) {
    sibling = siblings[++index];

    if (
      is('PunctuationNode', sibling) &&
      punctuation.test(toString(sibling))
    ) {
      sibling = siblings[++index];
    }

    if (is('WordNode', sibling)) {
      other = sibling;
    }
  }

  return other;
}

/** Classify a word. */
function classify(value) {
  var type = null;
  var normal;

  value = value.replace(digits, toWords).split(split, 1)[0];
  normal = lower(value);

  if (requiresA(value)) {
    type = 'a';
  }

  if (requiresAn(value)) {
    type = type === 'a' ? 'a-or-an' : 'an';
  }

  if (!type && normal === value) {
    type = vowel.test(normal.charAt(0)) ? 'an' : 'a';
  }

  return type;
}

/** Create a test based on a list of phrases. */
function factory(list) {
  var expressions = [];
  var sensitive = [];
  var insensitive = [];

  construct();

  return test;

  function construct() {
    var length = list.length;
    var index = -1;
    var value;
    var normal;

    while (++index < length) {
      value = list[index];
      normal = value === lower(value);

      if (value.charAt(value.length - 1) === '*') {
        /* Regexes are insensitive now, once we need them this
         * should check for `normal` as well. */
        expressions.push(new RegExp('^' + value.slice(0, -1), 'i'));
      } else {
        (normal ? insensitive : sensitive).push(value);
      }
    }
  }

  function test(value) {
    var normal = lower(value);
    var length;
    var index;

    if (
      sensitive.indexOf(value) !== -1 ||
      insensitive.indexOf(normal) !== -1
    ) {
      return true;
    }

    length = expressions.length;
    index = -1;

    while (++index < length) {
      if (expressions[index].test(value)) {
        return true;
      }
    }

    return false;
  }
}

/** Lower-case `value`. */
function lower(value) {
  return value.toLowerCase();
}
